/*
 *  bignum.c
 *  GlowWorm
 *
 *  Created by Curtis Jones on 2006.06.28.
 *  Copyright 2006 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

extern void bzero (void *, size_t);
extern void * memcpy (void *, const void *, size_t);
extern int printf(const char *, ...);
extern int snprintf(char *, size_t, const char *, ...);

char * BIGNUM_BASE_CHAR = "0123456789abcdefghijklmnopqrstuvwxyz";

static bignum_t * BIGNUM_ZERO;
static bignum_t * BIGNUM_ONE;
static bignum_t * BIGNUM_TWO;
static bignum_t * BIGNUM_THREE;
static bignum_t * BIGNUM_FOUR;
static bignum_t * BIGNUM_FIVE;
static bignum_t * BIGNUM_SIX;
static bignum_t * BIGNUM_SEVEN;
static bignum_t * BIGNUM_EIGHT;
static bignum_t * BIGNUM_NINE;
static bignum_t * BIGNUM_TEN;
static bignum_t * BIGNUM_ELEVEN;
static bignum_t * BIGNUM_TWELVE;
static bignum_t * BIGNUM_THIRTEEN;
static bignum_t * BIGNUM_FOURTEEN;
static bignum_t * BIGNUM_FIFTEEN;
static bignum_t * BIGNUM_SIXTEEN;
static bignum_t * BIGNUM_SEVENTEEN;
static bignum_t * BIGNUM_EIGHTEEN;
static bignum_t * BIGNUM_NINETEEN;
static bignum_t * BIGNUM_TWENTY;
static bignum_t * BIGNUM_THIRTY_TWO;
static bignum_t * BIGNUM_ONE_HUNDRED;





inline bignum_t * bignum_zero () { return BIGNUM_ZERO; }
inline bignum_t * bignum_one () { return BIGNUM_ONE; }
inline bignum_t * bignum_two () { return BIGNUM_TWO; }
inline bignum_t * bignum_three () { return BIGNUM_THREE; }
inline bignum_t * bignum_four () { return BIGNUM_FOUR; }
inline bignum_t * bignum_five () { return BIGNUM_FIVE; }
inline bignum_t * bignum_six () { return BIGNUM_SIX; }
inline bignum_t * bignum_seven () { return BIGNUM_SEVEN; }
inline bignum_t * bignum_eight () { return BIGNUM_EIGHT; }
inline bignum_t * bignum_nine () { return BIGNUM_NINE; }
inline bignum_t * bignum_ten () { return BIGNUM_TEN; }
inline bignum_t * bignum_eleven () { return BIGNUM_ELEVEN; }
inline bignum_t * bignum_twelve () { return BIGNUM_TWELVE; }
inline bignum_t * bignum_thirteen () { return BIGNUM_THIRTEEN; }
inline bignum_t * bignum_fourteen () { return BIGNUM_FOURTEEN; }
inline bignum_t * bignum_fifteen () { return BIGNUM_FIFTEEN; }
inline bignum_t * bignum_sixteen () { return BIGNUM_SIXTEEN; }
inline bignum_t * bignum_seventeen () { return BIGNUM_SEVENTEEN; }
inline bignum_t * bignum_eighteen () { return BIGNUM_EIGHTEEN; }
inline bignum_t * bignum_nineteen () { return BIGNUM_NINETEEN; }
inline bignum_t * bignum_twenty () { return BIGNUM_TWENTY; }
inline bignum_t * bignum_thirty_two () { return BIGNUM_THIRTY_TWO; }
inline bignum_t * bignum_one_hundred () { return BIGNUM_ONE_HUNDRED; }

/**
 * Allocates and returns a new bignum_t instance. It is initially set to 
 * positive zero. 
 *
 */
bignum_t *
bignum_alloc ()
{
	bignum_t * num = (bignum_t *)malloc(sizeof(bignum_t));
	
	BIGNUM_BASE_TYPE base = 0;
	
	bzero(num, sizeof(bignum_t));
	
	num->sign = BIGNUM_POSITIVE;
	num->oofm = 0;
	num->sofb = sizeof(BIGNUM_BASE_TYPE);
	num->size = 10;
	num->base = (u_int64_t)(base - 1) + 1;
	
	num->numb = (BIGNUM_BASE_TYPE *)malloc(10 * num->sofb);
	
	return num;
}

/**
 * Frees num along with any memory num had previously allocated.
 *
 */
void
bignum_free (bignum_t *num)
{
	if (num != NULL) {
		if (num->numb != NULL)
			free( num->numb );
		
		free( num );
	}
}

/**
 *
 *
 */
void
bignum_init ()
{
	BIGNUM_ZERO = bignum_alloc(); bignum_set(BIGNUM_ZERO, 0);
	BIGNUM_ONE = bignum_alloc(); bignum_set(BIGNUM_ONE, 1);
	BIGNUM_TWO = bignum_alloc(); bignum_set(BIGNUM_TWO, 2);
	BIGNUM_THREE = bignum_alloc(); bignum_set(BIGNUM_THREE, 3);
	BIGNUM_FOUR = bignum_alloc(); bignum_set(BIGNUM_FOUR, 4);
	BIGNUM_FIVE = bignum_alloc(); bignum_set(BIGNUM_FIVE, 5);
	BIGNUM_SIX = bignum_alloc(); bignum_set(BIGNUM_SIX, 6);
	BIGNUM_SEVEN = bignum_alloc(); bignum_set(BIGNUM_SEVEN, 7);
	BIGNUM_EIGHT = bignum_alloc(); bignum_set(BIGNUM_EIGHT, 8);
	BIGNUM_NINE = bignum_alloc(); bignum_set(BIGNUM_NINE, 9);
	BIGNUM_TEN = bignum_alloc(); bignum_set(BIGNUM_TEN, 10);
	BIGNUM_ELEVEN = bignum_alloc(); bignum_set(BIGNUM_ELEVEN, 11);
	BIGNUM_TWELVE = bignum_alloc(); bignum_set(BIGNUM_TWELVE, 12);
	BIGNUM_THIRTEEN = bignum_alloc(); bignum_set(BIGNUM_THIRTEEN, 13);
	BIGNUM_FOURTEEN = bignum_alloc(); bignum_set(BIGNUM_FOURTEEN, 14);
	BIGNUM_FIFTEEN = bignum_alloc(); bignum_set(BIGNUM_FIFTEEN, 15);
	BIGNUM_SIXTEEN = bignum_alloc(); bignum_set(BIGNUM_SIXTEEN, 16);
	BIGNUM_SEVENTEEN = bignum_alloc(); bignum_set(BIGNUM_SEVENTEEN, 17);
	BIGNUM_EIGHTEEN = bignum_alloc(); bignum_set(BIGNUM_EIGHTEEN, 18);
	BIGNUM_NINETEEN = bignum_alloc(); bignum_set(BIGNUM_NINETEEN, 19);
	BIGNUM_TWENTY = bignum_alloc(); bignum_set(BIGNUM_TWENTY, 20);
	BIGNUM_THIRTY_TWO = bignum_alloc(); bignum_set(BIGNUM_THIRTY_TWO, 32);
	BIGNUM_ONE_HUNDRED = bignum_alloc(); bignum_set(BIGNUM_ONE_HUNDRED, 100);
}

/**
 * Prints a base 16 representation of num.
 *
 */
void
bignum_print (bignum_t *num)
{
	u_int64_t i = 0;
	u_int64_t n = num->oofm;
	
	printf("bignum={sign=%d, oofm=%llu, size=%llu, numb=0x",
				 num->sign, num->oofm, num->size);
	
	if (n > 0) {
		for (i = n-1; i >= 0 && i < n; --i)
#if defined(BIGNUM_SOFB_08)
			printf("%02hhX ", num->numb[i]);
#elif defined(BIGNUM_SOFB_16)
			printf("%04hX ", num->numb[i]);
#elif defined(BIGNUM_SOFB_32)
			printf("%08lX ", num->numb[i]);
#elif defined(BIGNUM_SOFB_64)
			printf("%016llX ", num->numb[i]);
#endif
	}
	else
		printf("0");
	
	printf("}\n");
}

/**
 * Converts num to base 16 and puts that value into str. If len is insufficient
 * the value is truncated. Returns the total number of characters in the 
 * resulting string.
 *
 */
int
bignum_cstr (bignum_t *num, int len, char *str)
{
	u_int64_t i = 0;
	u_int64_t n = num->oofm;
	int pnt = 0;
	
	pnt += snprintf(str, len, "0x");
	
	if (n > 0) {
		for (i = n-1; i >= 0 && i < n; --i)
#if defined(BIGNUM_SOFB_08)
			pnt += snprintf(str+pnt, len-pnt, "%02hhX ",  num->numb[i]);
#elif defined(BIGNUM_SOFB_16)
			pnt += snprintf(str+pnt, len-pnt, "%04hX ",   num->numb[i]);
#elif defined(BIGNUM_SOFB_32)
			pnt += snprintf(str+pnt, len-pnt, "%08lX ",   num->numb[i]);
#elif defined(BIGNUM_SOFB_64)
			pnt += snprintf(str+pnt, len-pnt, "%016llX ", num->numb[i]);
#endif
	}
	else
		pnt += snprintf(str+pnt, len-pnt, "0 ");
	
	return pnt;
}

/**
 * Don't forget to free the returned char*.
 *
 */
char *
bignum_cstr_base2 (bignum_t *num, int precision, int base)
{
	char * str;
	
	if (NULL == (str = (char *)malloc(precision+1)))
		return NULL;
	
	bignum_cstr_base(num, precision, str, base);
	
	return str;
}

/**
 * Converts num into the specified base and puts that value into str. If 
 * str_len is less than the necessary length to hold the new number, then
 * scientific notation is used. Returns the total number of digits that the
 * converted number *should* be represented by if str_len allows for it.
 *
 * The efficiency of this could be improved by allocating an additional buffer
 * to use as a temporary storage space and then just memmove()
 */
int
bignum_cstr_base (bignum_t *num, int str_len, char *str, int base)
{
	if (base < 2)
		return -1;
	
	if (base > 36)
		return -2;
	
	int str_pos = 0;
	char * str_ptr = str + (str_len - 2);
	
	bignum_t * num_base = bignum_alloc();					// target base
	bignum_t * num_tmp1 = bignum_alloc();					// quotient from division operation
	bignum_t * num_tmp2 = bignum_alloc();					// remainder from division operation
	bignum_t * num_tmp3 = bignum_alloc();					// working copy of num
	bignum_t * num_tmp4 = NULL;										// used for swapping tmp1 and tmp3
	
	bignum_set(num_base, base);
	bignum_set_bn(num_tmp3, num);
	
	// if the number in zero, then the result is the same for all bases, so just
	// set it and return.
	//
	if (bignum_iszero(num)) {
		str[0] = '0';
		str[1] = '\0';
		return 1;
	}
	
	// divide the given number by the given radix. the quotient of each such
	// division operation represents the next largest magnitude of the base
	// converted value. stop when we hit zero.
	//
	while (!bignum_iszero(num_tmp3))
	{
		bignum_div(num_tmp1, num_tmp2, num_tmp3, num_base);
		
		*(str_ptr-((str_pos++)%(str_len-1))) = BIGNUM_BASE_CHAR[__bignum_get(num_tmp2,0)];
		
		num_tmp4 = num_tmp3;
		num_tmp3 = num_tmp1;
		num_tmp1 = num_tmp4;
	}
	
	// set the terminating binary zero at the end of the string.
	//
	*(str_ptr+1) = '\0';
	
	// if the resulting string is less than the size of the string buffer we were
	// given, then shift the string to the front of the buffer and set a 
	// terminating zero after it.
	//
	if (str_pos < str_len - 1) {
		memmove(str, str+(str_len-str_pos-1), str_pos);
		*(str+str_pos) = '\0';
	}
	
	// if the resulting string is greater than the size of the string buffer we 
	// were given, then we need to shift the numbers around so that they start at 
	// the top of the string buffer.
	//
	// the method used here involves a read position and a write position. the 
	// read position is set at the point where the last (ie, highest order of 
	// magnitude) value was written, and the write position is set to the start 
	// of the same buffer.
	//
	// digit values are swapped, amounting to no more than one swap per digit in
	// the string and possibly fewer.
	//
	// this process is complicated a little bit by the need to insert a decimal 
	// point after the first digit because we need to express the value in 
	// scientific notation. this is accomplished by manually setting the the 
	// first digit and the following decimal point and setting the read position
	// to minus-one the position of the highest order of magintude value (which,
	// of course, is the position of the lowest order of magnitude value), which
	// is where the second digit in the un-altered string - which would have been
	// over-written by the decimal point - is stored.
	//
	// finally, there is special case handling for situations where no shifting
	// is required or the digits only need to be shifted once.
	//
	else if (str_pos > str_len - 1)
	{
		char tmp;
		char expo[10];
		int read = ((str_len-1) - (str_pos % (str_len-1))) - 1;
		int write = 0;
		int elen = 0;
		
		elen = snprintf(expo, 10, "x10^%d", str_pos-1);
		
		// if the digits are already in the correct oreientation, then we just need
		// to insert the decimal point after shifting down all but the first digit.
		//
		if (read == -1 || read == str_len - 2) {
			memmove(str+2, str+1, str_len-3);
			*(str+1) = '.';
		}
		
		// if the digits are shifted by one, then "manually" set the first digit
		// and the decimal point and we're done.
		//
		else if (read == 0) {
			*(str+0) = *(str+1);
			*(str+1) = '.';
		}
		
		// some arbitrary amount of shifting needs to take place and room needs to
		// be made for the decimal point. this'll do all of that. and more!
		//
		else {
			*(str+(read++)) = *(str+write);
			*(str+(write++)) = *(str+read);
			*(str+(read++)) = *(str+write);
			*(str+(write++)) = '.';
			
			while (1) {
				if (read >= str_len-1) {
					read = ((str_len-1) - (str_pos % (str_len-1))) - 1;
					
					if (read <= write)
						break;
				}
				
				// if the write position is now over-lapping with the scientific
				// notation portion of the string, then there's no point in continuing.
				//
				if (write >= str_len - elen)
					break;
				
				tmp = *(str+write);
				*(str+write) = *(str+read);
				*(str+read) = tmp;
				
				read++;
				write++;
			}
		}
		
		// copy the scientific notation text onto the end of the string.
		//
		memcpy(str+(str_len-(elen+1)), expo, elen);
	}
	
	bignum_free(num_base);
	bignum_free(num_tmp1);
	bignum_free(num_tmp2);
	bignum_free(num_tmp3);
	
	return str_pos;
}

/**
 * Converts num into the specified base and puts that value into str. If 
 * str_len is less than the necessary length to hold the new number, then
 * scientific notation is used. Returns the total number of digits that the
 * converted number *should* be represented by if str_len allows for it.
 *
 * Same as bignum_cstr_base, but incurs additional memory overhead for the
 * sake of speed.
 *
 */
/*
int
bignum_cstr_base_fast (bignum_t *num, int str_len, char *str, int base)
{
	if (base < 2)
		return -1;
	
	if (base > 36)
		return -2;
	
	int str_pos = 0;
	char * str_ptr = str + (str_len - 2);
	
	bignum_t * num_base = bignum_alloc();					// target base
	bignum_t * num_tmp1 = bignum_alloc();					// quotient from division operation
	bignum_t * num_tmp2 = bignum_alloc();					// remainder from division operation
	bignum_t * num_tmp3 = bignum_alloc();					// working copy of num
	bignum_t * num_tmp4 = NULL;										// used for swapping tmp1 and tmp3
	
	bignum_set(num_base, base);
	bignum_set_bn(num_tmp3, num);
	
	// if the number in zero, then the result is the same for all bases, so just
	// set it and return.
	//
	if (bignum_iszero(num)) {
		str[0] = '0';
		str[1] = '\0';
		return 1;
	}
	
	// divide the given number by the given radix. the quotient of each such
	// division operation represents the next largest magnitude of the base
	// converted value. stop when we hit zero.
	//
	while (!bignum_iszero(num_tmp3))
	{
		bignum_div(num_tmp1, num_tmp2, num_tmp3, num_base);
		
		*(str_ptr-((str_pos++)%(str_len-1))) = BIGNUM_BASE_CHAR[__bignum_get(num_tmp2,0)];
		
		num_tmp4 = num_tmp3;
		num_tmp3 = num_tmp1;
		num_tmp1 = num_tmp4;
	}
	
	// set the terminating binary zero at the end of the string.
	//
	*(str_ptr+1) = '\0';
	
	// if the resulting string is less than the size of the string buffer we were
	// given, then shift the string to the front of the buffer and set a 
	// terminating zero after it.
	//
	if (str_pos < str_len - 1) {
		memmove(str, str+(str_len-str_pos-1), str_pos);
		*(str+str_pos) = '\0';
	}
	
	// if the resulting string is greater than the size of the string buffer we 
	// were given, then we need to shift the numbers around so that they start at 
	// the top of the string buffer.
	//
	// the method used here involves a read position and a write position. the 
	// read position is set at the point where the last (ie, highest order of 
	// magnitude) value was written, and the write position is set to the start 
	// of the same buffer.
	//
	// digit values are swapped, amounting to no more than one swap per digit in
	// the string and possibly fewer.
	//
	// this process is complicated a little bit by the need to insert a decimal 
	// point after the first digit because we need to express the value in 
	// scientific notation. this is accomplished by manually setting the the 
	// first digit and the following decimal point and setting the read position
	// to minus-one the position of the highest order of magintude value (which,
	// of course, is the position of the lowest order of magnitude value), which
	// is where the second digit in the un-altered string - which would have been
	// over-written by the decimal point - is stored.
	//
	// finally, there is special case handling for situations where no shifting
	// is required or the digits only need to be shifted once.
	//
	else if (str_pos > str_len - 1)
	{
		char tmp;
		char expo[10];
		int read = ((str_len-1) - (str_pos % (str_len-1))) - 1;
		int write = 0;
		int elen = 0;
		
		elen = snprintf(expo, 10, "x10^%d", str_pos-1);
		
		// if the digits are already in the correct oreientation, then we just need
		// to insert the decimal point after shifting down all but the first digit.
		//
		if (read == -1 || read == str_len - 2) {
			memmove(str+2, str+1, str_len-3);
			*(str+1) = '.';
		}
		
		// if the digits are shifted by one, then "manually" set the first digit
		// and the decimal point and we're done.
		//
		else if (read == 0) {
			*(str+0) = *(str+1);
			*(str+1) = '.';
		}
		
		// some arbitrary amount of shifting needs to take place and room needs to
		// be made for the decimal point. this'll do all of that. and more!
		//
		else {
			char * buf;
			
			if (NULL == (buf = malloc(strlen)))
				return -1;
			
			memcpy(buf, str+read, strlen);
			
			*(str+(read++)) = *(str+write);
			*(str+(write++)) = *(str+read);
			*(str+(read++)) = *(str+write);
			*(str+(write++)) = '.';
			
			while (1) {
				if (read >= str_len-1) {
					read = ((str_len-1) - (str_pos % (str_len-1))) - 1;
					
					if (read <= write)
						break;
				}
				
				// if the write position is now over-lapping with the scientific
				// notation portion of the string, then there's no point in continuing.
				//
				if (write >= str_len - elen)
					break;
				
				tmp = *(str+write);
				*(str+write) = *(str+read);
				*(str+read) = tmp;
				
				read++;
				write++;
			}
		}
		
		// copy the scientific notation text onto the end of the string.
		//
		memcpy(str+(str_len-(elen+1)), expo, elen);
	}
	
	bignum_free(num_base);
	bignum_free(num_tmp1);
	bignum_free(num_tmp2);
	bignum_free(num_tmp3);
	
	return str_pos;
}
*/

/**
 * Sets the value of num to the value contained in buf. Returns an instance
 * of num.
 *
 */
bignum_t *
bignum_setbytes (bignum_t *num, unsigned char *buf, int len)
{
	int i = 0;
	u_int64_t pos = 0;
	BIGNUM_BASE_TYPE val = 0;
	
	__bignum_zero(num);
	
	for (i = 0; i <= len - 4; i+=4) {
		val  = 0;
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+4)]) << 24);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+3)]) << 16);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+2)]) <<  8);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+1)])      );
		
		__bignum_set(num, pos++, val);
	}
	
	if (len - i == 3) {
		val  = 0;
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+3)]) << 16);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+2)]) <<  8);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+1)])      );
		
		__bignum_set(num, pos++, val);
	}
	else if (len - i == 2) {
		val  = 0;
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+2)]) <<  8);
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+1)])      );
		
		__bignum_set(num, pos++, val);
	}
	else if (len - i == 1) {
		val  = 0;
		val |= (((BIGNUM_BASE_TYPE)buf[len-(i+1)])      );
		
		__bignum_set(num, pos++, val);
	}
	
	num->oofm = pos;
	
	return num;
}

/**
 * Copies the value of num into buf. Returns buf.
 *
 */
unsigned char *
bignum_getbytes (bignum_t *num, unsigned char *buf)
{
	u_int64_t i=1, j=0, oofm=num->oofm;
	
	if (oofm == 0) {
		*buf = '\0';
		return buf;
	}
	
	if (num->numb[oofm-1] > 0x00FF0000ull) {
		buf[j++] = (num->numb[oofm-1] >> 24) & 0xFF;
		buf[j++] = (num->numb[oofm-1] >> 16) & 0xFF;
		buf[j++] = (num->numb[oofm-1] >>  8) & 0xFF;
		buf[j++] = (num->numb[oofm-1]      ) & 0xFF;
	}
	else if (num->numb[oofm-1] > 0x0000FF00ull) {
		buf[j++] = (num->numb[oofm-1] >> 16) & 0xFF;
		buf[j++] = (num->numb[oofm-1] >>  8) & 0xFF;
		buf[j++] = (num->numb[oofm-1]      ) & 0xFF;
	}
	else if (num->numb[oofm-1] > 0x000000FFull) {
		buf[j++] = (num->numb[oofm-1] >>  8) & 0xFF;
		buf[j++] = (num->numb[oofm-1]      ) & 0xFF;
	}
	else
		buf[j++] = (num->numb[oofm-1]      ) & 0xFF;
		
	for (oofm -= 2; i < num->oofm; ++i, oofm--) {
		buf[j++] = (num->numb[oofm] >> 24) & 0xFF;
		buf[j++] = (num->numb[oofm] >> 16) & 0xFF;
		buf[j++] = (num->numb[oofm] >>  8) & 0xFF;
		buf[j++] = (num->numb[oofm]      ) & 0xFF;
	}
	
	return buf;
}

/**
 * Returns a u_int32_t representation of the bignum_t. If the bignum_t 
 * overflows the u_int32_t type, that data is discarded, so make sure you are
 * aware of whether the bignum_t is going to fit into this data size.
 *
 */
u_int32_t
bignum_get_uint32 (bignum_t *num)
{
	if (num == NULL)
		return 0;
	
	return num->numb[0];
}

/**
 * Shifts the value of num by cnt.
 *
 */
bignum_t * 
bignum_rshift (bignum_t *num, u_int64_t cnt)
{
	u_int64_t i;																// counter
	u_int64_t bpmag = num->sofb * 8;						// bits per magnitude
	u_int64_t large = cnt / bpmag;							// 
	u_int64_t small = cnt - (large * bpmag);		// 
	
	for (i = 0; i < num->oofm - 1; ++i) {
		num->numb[i] >>= small;
		num->numb[i] |= (num->numb[i+1] << (bpmag - small));
	}
	
	num->numb[i] >>= small;
	
	return num;
}

/**
 * Sets the value of res to num right-shifted amt times. If res is null, a new
 * bignum_t is allocated.
 *
 * Returns res, or the newly allocated res. Returns NULL if num or amt are 
 * null.
 *
 */
bignum_t *
bignum_rshift_multi (bignum_t *res, bignum_t *num, bignum_t *amt)
{
	u_int32_t oofm_cnt = 0;
	u_int32_t bits_cnt = 0;
	
	if (num == NULL)
		return NULL;
	
	if (amt == NULL)
		return NULL;
	
	if (res == NULL)
		if (NULL == (res = bignum_alloc()))
			return NULL;
	
	__bignum_zero(res);
	
	// if amt is greater than num->oofm * 32 then
	//   set res = 0
	//
	// 
	
	// u_int64_t oofm = num->oofm * 
	
	// memset
	
	// use memset for the "(int) oofm / 32" amount of the shift, and
	// use >> for the "oofm % 32" amount of the shift
	
	// determine the number of 32-bit orders-of-magnitude that we need to shift
	// the value by, which can be accomplished with memset, and also determine
	// the number of bits that need to be shifted, using >>.
	{
		bignum_t * shift_oofm_cnt = bignum_alloc();
		bignum_t * shift_bits_cnt = bignum_alloc();
		
		bignum_div(shift_oofm_cnt, shift_bits_cnt, amt, bignum_thirty_two());
		
		oofm_cnt = bignum_get_uint32(shift_oofm_cnt);
		bits_cnt = bignum_get_uint32(shift_bits_cnt);
		
		bignum_free(shift_oofm_cnt);
		bignum_free(shift_bits_cnt);
	}
	
	
	
	return res;
}

/**
 * Returns 1 if the bit specified by bit is set and zero otherwise.
 *
 */
int
bignum_isbitset (bignum_t *num, u_int64_t bit)
{
	u_int64_t bpmag = num->sofb * 8;						// bits per magnitude
	u_int64_t large = bit / bpmag;							// magnitude
	u_int64_t small = bit - (large * bpmag);		// bit within magnitude
	
	return (num->numb[large] >> (small - 1)) & 0x1;
}

/**
 * Returns the lowest set bit of num or negative one if num is zero.
 *
 */
u_int64_t
bignum_lowest_set_bit (bignum_t *num)
{
	u_int64_t i, j;
	u_int64_t bpmag = num->sofb * 8;
	
	for (i = 0; i < num->oofm; ++i) {
		for (j = 0; j < bpmag; ++j) {
			if ((num->numb[i] >> j) & 1)
				return (i * bpmag) + j;
		}
	}
	
	return -1;
}

/**
 * Returns the number of bits in num.
 *
 */
u_int64_t
bignum_bitlen (bignum_t *num)
{
	int i = 0;
	u_int64_t bpmag = num->sofb * 8;
	
	if (bignum_iszero(num))
		return 0;
	
	for (i = bpmag - 1; i >= 0; --i)
		if ((num->numb[num->oofm-1] >> i) & 1)
			return ((num->oofm-1) * bpmag) + (i+1);
	
	return 0;
}

/**
 * Returns the number of bytes in num.
 *
 */
u_int64_t
bignum_bytelen (bignum_t *num)
{
	int bytes = 0;
	BIGNUM_BASE_TYPE numb = num->numb[num->oofm - 1];
	
	if (num->oofm == 0)
		return 0;
	
	else if (num->oofm > 1)
		bytes = (num->oofm - 1) * num->sofb;
	
	if ((numb >> 24) & 0xFF)
		bytes += 4;
	else if ((numb >> 16) & 0xFF)
		bytes += 3;
	else if ((numb >>  8) & 0xFF)
		bytes += 2;
	else if ((numb      ) & 0xFF)
		bytes += 1;
	
	return bytes;
}





/**
 * Returns the BIGNUM_BASE_TYPE of num at position pos.
 *
 */
inline BIGNUM_BASE_TYPE
__bignum_get (bignum_t *num, u_int64_t pos)
{
	if (pos >= num->oofm)
		return 0;
	
	return *(num->numb+pos);
}

/**
 * Sets the BIGNUM_BASE_TYPE of num at position pos to val.
 *
 */
inline void
__bignum_set (bignum_t *num, u_int64_t pos, BIGNUM_BASE_TYPE val)
{
	if (pos >= BIGNUM_MAX_ORDERS)
		return;
	
	if (num->size <= pos)
		__bignum_grow(num, pos+1);
	
	if (num->size <= pos)
		printf("%s().. failed to grow to %d\n", __PRETTY_FUNCTION__, (pos+1));
	else
		num->numb[pos] = val;
}

/**
 * Increases the size of num to size orders of magnitude (BIGNUM_BASE_TYPE),
 * if necessary. Does not shrink the size of num if it is greater than size.
 *
 */
void
__bignum_grow (bignum_t *num, u_int64_t size)
{
	BIGNUM_BASE_TYPE * ptr = NULL;
	
	if (size >= BIGNUM_MAX_ORDERS) {
		printf("%s().. requested size of %llu exceeds max of %d\n", __PRETTY_FUNCTION__, size, BIGNUM_MAX_ORDERS);
		return;
	}
	
	if (size < num->size)
		return;
	
	if (num->size * 2 >= size)
		size = BIGNUM_MIN(num->size*2, BIGNUM_MAX_ORDERS);
	else
		size = BIGNUM_MIN(size*2, BIGNUM_MAX_ORDERS);
	
	if (NULL == (ptr = (BIGNUM_BASE_TYPE *)realloc(num->numb, size * num->sofb))) {
		printf("%s().. failed to realloc, %d\n", __PRETTY_FUNCTION__, errno);
		return;
	}
	
	num->numb = ptr;
	num->size = size;
}

/**
 * Set the value of num to zero.
 *
 */
void
__bignum_zero (bignum_t *num)
{
	bzero(num->numb, num->size * num->sofb);
	num->oofm = 0;
	num->sign = BIGNUM_POSITIVE;
}

/**
 * Removes any trailing zeroes from num.
 *
 */
void
__bignum_trim (bignum_t *num)
{
	u_int64_t i = 0;
	
	for (i = num->oofm - 1; i < ((u_int64_t)0) - 1 && num->oofm > 0; --i) {
		if (num->numb[i] != 0)
			break;
		else
			num->oofm--;
	}
}
