/*
 *  bignum_cmp.c
 *  BigMath
 *
 *  Created by Curtis Jones on 2007.08.05.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

/**
 * return zero if the two numbers are the same
 * returns greater than zero if the first number is greater than the second number
 * returns less than zero if the first number is less than the second number
 *
 */
inline int
bignum_cmp (bignum_t *num1, bignum_t *num2)
{
	u_int64_t i = 0;
	u_int64_t num1_i=0, num2_i=0;
	u_int64_t num1_start = 0;
	u_int64_t num2_start = 0;
	int num1_found = FALSE;
	int num2_found = FALSE;
	
	if (num1 == NULL)
		return 0;
	
	if (num2 == NULL)
		return 0;
	
	if (num1->oofm == 0 && num2->oofm == 0)
		return 0;
	
	if (num1->sign == BIGNUM_NEGATIVE && num2->sign == BIGNUM_POSITIVE)
		return -1;
	else if (num2->sign == BIGNUM_NEGATIVE && num1->sign == BIGNUM_POSITIVE)
		return 1;
	
	for (i = 0; i < num1->oofm; ++i) {
		if (num1->numb[i] != 0) {
			num1_start = i;
			num1_found = TRUE;
		}
	}
	
	for (i = 0; i < num2->oofm; ++i) {
		if (num2->numb[i] != 0) {
			num2_start = i;
			num2_found = TRUE;
		}
	}
	
	if (num1_found == FALSE && num2_found == FALSE)
		return 0;
	else if (num1_found == FALSE)
		return -1;
	else if (num2_found == FALSE)
		return 1;
	
	if (num1_start != num2_start)
		return num1_start > num2_start ? 1 : -1;
	
	for (num1_i=num1_start, num2_i=num2_start; num1_i <= num1_start; --num1_i, --num2_i) {
		if (num1->numb[num1_i] > num2->numb[num2_i])
			return 1;
		else if (num1->numb[num1_i] < num2->numb[num2_i])
			return -1;
	}
	
	return 0;
}

/**
 * Returns 1 if num1 and num2 are equal; zero otherwise.
 *
 */
inline int
bignum_eq (bignum_t *num1, bignum_t *num2)
{
	return bignum_cmp(num1, num2) == 0 ? TRUE : FALSE;
}

/**
 * Returns 1 if num1 is greater than num2; zero otherwise.
 *
 */
inline int
bignum_gt (bignum_t *num1, bignum_t *num2)
{
	return bignum_cmp(num1, num2) > 0 ? TRUE : FALSE;
}

/**
 * Return 1 if num1 is less than num2; zero otherwise.
 *
 */
inline int
bignum_lt (bignum_t *num1, bignum_t *num2)
{
	return bignum_cmp(num1, num2) < 0 ? TRUE : FALSE;
}

/**
 * Return 1 if num is equal to zero; zero otherwise.
 *
 */
inline int
bignum_iszero (bignum_t *num)
{
	if (num->oofm == 0 || (num->oofm == 1 && num->numb[0] == 0))
		return TRUE;
	else
		return FALSE;
}
