/*
 *  bignum_pow.c
 *  BigMath
 *
 *  Created by Curtis Jones on 2007.08.05.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "bignum.h"

// Knuth 4.6.3 - Algorithm A
//
bignum_t *
bignum_pow (bignum_t *Y, bignum_t *num, bignum_t *exp)
{
	bignum_t *N = bignum_alloc();
	bignum_t *Z = bignum_alloc();
	bignum_t *t = bignum_alloc();
	bignum_t *x = bignum_alloc();
	bignum_t *w = bignum_alloc();
	
	bignum_set_bn(N, exp);
	bignum_set(Y, 1);
	bignum_set_bn(Z, num);
	bignum_set(w, 2);
	
	while (1) {
		bignum_mod(t, N, w);							// t = N % w
		bignum_div(x, NULL, N, w);				// x = N / w
		bignum_set_bn(N, x);							// N = x
		
		if (!bignum_iszero(t)) {
			bignum_mul(x, Z, Y);						// x = Z * Y
			bignum_set_bn(Y, x);						// Y = x
			
			if (bignum_iszero(N))						// N == 0
				break;
		}
		
		bignum_mul(x, Z, Z);							// x = Z * Z
		bignum_set_bn(Z, x);							// Z = x
	}
	
	bignum_free( N );
	bignum_free( Z );
	bignum_free( t );
	bignum_free( x );
	bignum_free( w );
	
	return Y;
}
